"""'lit' Testing Tool"""

__author__ = 'Daniel Dunbar'
__email__ = 'daniel@minormatter.com'
__versioninfo__ = (0, 9, 1)
__version__ = '.'.join(str(v) for v in __versioninfo__) + 'dev'

__all__ = []

from .main import main
